/*
 * init/x86.c
 *
 * Copyright (C) 2000 Russell King.
 *
 * This file provides the glue for the X86 emulator.
 */
#include <bios/types.h>
#include <bios/stdio.h>
#include <bios/string.h>
#include <bios/malloc.h>
#include <bios/pci.h>
#include <bios/x86.h>
#include <x86emu.h>

/*
 * The x86 memory map consists of 16 64K "pages".  This
 * allows us to point x86 addresses 0xa0000 - 0xbf000
 * at the PCI memory space.
 */
#define NR_PAGES	16

static unsigned char *pages[NR_PAGES];


/*
 * i8253 emulation.  We may not have a PCI Southbridge.  In fact, if we
 * do have one, we don't really want to touch it.  Instead, we emulate
 * this beast.
 */
struct i8253_timer {
	u32 latch;
	u32 ctr;
	u8  div_low, div_high;
	u8  mode, high;
};

static struct i8253_timer i8253_timer[3];

static unsigned int i8253_read(unsigned int ctr)
{
	struct i8253_timer *t = i8253_timer + ctr;
	unsigned int mode = t->mode & 0x30;

	if (mode) {
		unsigned int val, high;

		val = t->latch;

		high = t->high;
		if (high)
			val >>= 8;

		if (mode == 0x30)
			t->high = high ^ 1;

		return val;
	}
	return 0;
}

static void i8253_write(unsigned int ctr, unsigned int val)
{
	struct i8253_timer *t = i8253_timer + ctr;
	unsigned int mode = t->mode & 0x30;

	if (mode) {
		unsigned int high;

		high = t->high;
		if (high)
			t->div_high = val;
		else
			t->div_low = val;

		if (mode == 0x30)
			t->high = high ^ 1;
		t->ctr = t->div_low | val << 16;
	}
}

static void i8253_write_ctrl(unsigned int val)
{
	if ((val & 0xc0) != 0xc0) {
		struct i8253_timer *t = i8253_timer + (val >> 6);
		unsigned int mode = val & 0x30;

		t->high = mode == 0x10 ? 1 : 0;

		if (mode)
			t->mode = val;
		else
			t->latch = t->ctr;
	} else {
		printf("x86: unsupported timer command %02x\n", val);
	}
}

/*
 * Emulate one tick of the counter
 */
static void i8253_cycle(void)
{
	int i;

	for (i = 0; i < 3; i++) {
		struct i8253_timer *t = i8253_timer + i;
		unsigned int dec, msk;

		if ((t->mode & 0x0e) == 0x06) {
			dec = 2;
			msk = ~1;
		} else {
			dec = 1;
			msk = ~0;
		}
		t->ctr -= dec;
		if ((t->ctr & msk) == 0)
			t->ctr = t->div_low | t->div_high << 8;
	}
}

static void x86_cycle(void)
{
	i8253_cycle();
}

/*
 * x86 port read/write support
 */

static u8 x86_inb(u16 addr)
{
	unsigned int val;

	if (addr >= 0x3b0 && addr < 0x3e0) {
		val = pci_io_read_byte(addr);
		goto out;
	}
	if (addr >= 0x40 && addr < 0x43) {	/* 8253 timer */
		val = i8253_read(addr - 0x40);
		goto out;
	}
	val = 0;
	if (addr == 0x61 || addr == 0x80)
		goto out;

	printf("x86: inb: unknown port 0x%04x\n", addr);

out:	x86_cycle();
	return val;
}

static u16 x86_inw(u16 addr)
{
	unsigned int val;

	if (addr >= 0x3b0 && addr < 0x3e0) {
		val = pci_io_read_word(addr);
		goto out;
	}

	printf("x86: inw: unknown port 0x%04x\n", addr);
	val = 0;

out:	x86_cycle();
	x86_cycle();
	return val;
}

static u32 x86_inl(u16 addr)
{
	printf("x86: inl: unknown port 0x%04x\n", addr);

	x86_cycle();
	x86_cycle();
	x86_cycle();
	x86_cycle();
	return 0;
}

static void x86_out(u16 addr, unsigned int val, int sz)
{
	switch (addr) {
	case 0x3b0 ... 0x3df:	/* VGA */
		switch (sz) {
		case 1: pci_io_write_byte(val,addr); break;
		case 2: pci_io_write_word(val,addr); break;
		case 4: pci_io_write_long(val,addr); break;
		}
		break;

	case 0x61:
		break;

	case 0x43:		/* 8253 timer */
		if (sz == 1)
			i8253_write_ctrl(val);
		break;

	case 0x40 ... 0x42:	/* 8253 timer */
		if (sz == 1)
			i8253_write(addr - 0x40, val);
		break;

	default:
		printf("x86: out size %d to unknown port %04x\n", sz, addr);
	}
}

static void x86_outb(u16 addr, u8 val)
{
	x86_out(addr, val, 1);
	x86_cycle();
}

static void x86_outw(u16 addr, u16 val)
{
	x86_out(addr, val, 2);
	x86_cycle();
	x86_cycle();
}

static void x86_outl(u16 addr, u32 val)
{
	x86_out(addr, val, 4);
	x86_cycle();
	x86_cycle();
	x86_cycle();
	x86_cycle();
}

static const X86EMU_pioFuncs piofuncs = {
	inb:	x86_inb,
	inw:	x86_inw,
	inl:	x86_inl,
	outb:	x86_outb,
	outw:	x86_outw,
	outl:	x86_outl,
};

/*
 * x86 memory read/write support
 */

static void *x86_mem_bad(u32 addr, const char *fn)
{
	printf("x86: %s: address 0x%x out of range\n", fn, addr);
	HALT_SYS();
	return NULL;
}

static void *x86_mem(u32 addr, const char *fn)
{
	u32 page = addr >> 16;
	if (page < NR_PAGES && pages[page])
		return pages[page] + addr - (page << 16);

	return x86_mem_bad(addr, fn);
}

static u8 x86_rdb(u32 addr)
{
	void *mem;
	u32 val;

	x86_cycle();

	mem = x86_mem(addr, "rdb");

	__asm__(
	"teq	%1, #0
	ldrneb	%0, [%1]" :
	"=r" (val) : "0" (mem));

	return val;
}

static u16 x86_rdw(u32 addr)
{
	void *mem;
	u32 val1, val2;

	if (addr & 1)
		x86_cycle();
	x86_cycle();

	mem  = x86_mem(addr, "rdw");

	__asm__(
	"teq	%2, #0
	beq	1f
	tst	%2, #1
	ldrneb	%1, [%2, #1]
	ldrneb	%0, [%2]
	ldreqh	%0, [%2]
	orrne	%0, %0, %1, lsl #8
1:	" :
	"=r" (val1), "=r" (val2) : "0" (mem) : "cc");

	return val1;
}

static u32 x86_rdl(u32 addr)
{
	void *mem;
	u32 val1, val2, val3;

	if (addr & 1)
		x86_cycle();
	x86_cycle();
	x86_cycle();

	mem = x86_mem(addr, "rdl");

	__asm__(
	"teq	%3, #0
	beq	1f
	tst	%3, #3
	ldreq	%0, [%3]
	beq	1f
	ldrb	%1, [%3]
	ldrb	%2, [%3, #1]
	orr	%1, %1, %2, lsl #8
	ldrb	%2, [%3, #2]
	orr	%1, %1, %2, lsl #16
	ldrb	%2, [%3, #3]
	orr	%0, %1, %2, lsl #24
1:	" :
	"=r" (val1), "=r" (val2), "=r" (val3) : "0" (mem) : "cc");

	return val1;
}

static void x86_wrb(u32 addr, u8 val)
{
	void *mem = x86_mem(addr, "wrb");

	x86_cycle();

	if (mem)
		*(u8 *)mem = val;

	/*
	 * If we are writing into the interrupt area,
	 * then clear out our "catch all" version.
	 */
	if (mem && addr < 0x400)
		_X86EMU_intrTab[addr >> 2] = NULL;
}

static void x86_wrw(u32 addr, u16 val)
{
	void *mem = x86_mem(addr, "wrw");

	if (addr & 1)
		x86_cycle();
	x86_cycle();

	if (mem) {
		if (addr & 1) {
			((u8 *)mem)[0] = val;
			((u8 *)mem)[1] = val >> 8;
		} else {
			__asm__("str%?h\t%0,[%1]" : : "r" (val), "r" (mem));
		}
	}
	/*
	 * If we are writing into the interrupt area,
	 * then clear out our "catch all" version.
	 */
	if (mem && addr < 0x400)
		_X86EMU_intrTab[addr >> 2] = NULL;
}

static void x86_wrl(u32 addr, u32 val)
{
	void *mem = x86_mem(addr, "wrl");

	if (addr & 1)
		x86_cycle();
	x86_cycle();
	x86_cycle();

	if (mem) {
		if (addr & 3) {
			((u8 *)mem)[0] = val;
			((u8 *)mem)[1] = val >> 8;
			((u8 *)mem)[2] = val >> 16;
			((u8 *)mem)[3] = val >> 24;
		} else {
			*(u32 *)mem = val;
		}
	}
	/*
	 * If we are writing into the interrupt area,
	 * then clear out our "catch all" version.
	 */
	if (mem && addr < 0x400)
		_X86EMU_intrTab[addr >> 2] = NULL;
}

static const X86EMU_memFuncs memfuncs = {
	rdb:	x86_rdb,
	rdw:	x86_rdw,
	rdl:	x86_rdl,
	wrb:	x86_wrb,
	wrw:	x86_wrw,
	wrl:	x86_wrl,
};

/*
 * Don't fault on unknown interrupts - flag them!
 */
static void x86_dummy_intr(int nr)
{
	printf("x86: unknown interrupt 0x%02x, AX=%04x BX=%04x CX=%04x\n",
		nr, M.x86.R_AX, M.x86.R_BX, M.x86.R_CX);
}

/*
 * Copy into the x86 BIOS area
 */
void x86_copy_in(unsigned int addr, void *data, int sz)
{
	void *ptr;

	if (addr < 0xc0000 || addr + sz >= 0xdffff) {
		printf("x86: copy_in: bad address 0x%x\n", addr);
		return;
	}

	ptr = pages[12] + (addr - 0xc0000);
	memcpy(ptr, data, sz);
}

/*
 * Copy out of the x86 BIOS area
 */
void x86_copy_out(void *data, unsigned int addr, int sz)
{
	void *ptr;

	if (addr < 0xc0000 || addr + sz >= 0xdffff) {
		printf("x86: copy_out: bad address 0x%x\n", addr);
		return;
	}

	ptr = pages[12] + (addr - 0xc0000);
	memcpy(data, ptr, sz);
}

/*
 * Execute some x86 code
 */
void x86_call(unsigned int eip, unsigned int ax)
{
	_X86EMU_env.x86.R_AX = ax;
	_X86EMU_env.x86.R_IP = eip & 15;
	_X86EMU_env.x86.R_CS = eip >> 4;
	_X86EMU_env.x86.R_DS = eip >> 4;

	X86EMU_exec();
}

/*
 * Initialise the X86 core
 */
void x86_init(void)
{
	int i;

	/*
	 * Setup our memory I/O and port I/O functions
	 */
	X86EMU_setupPioFuncs((X86EMU_pioFuncs *)&piofuncs);
	X86EMU_setupMemFuncs((X86EMU_memFuncs *)&memfuncs);

	/*
	 * Initialise the stack segment
	 */
	_X86EMU_env.x86.R_SS = 0x0040;

	/*
	 * Make all interrupts point at our dummy int routine.
	 * As the bioses initialise, they will register their
	 * owne, and the entry in this table will be NULL'd out.
	 */
	for (i = 0; i < 256; i++)
		_X86EMU_intrTab[i] = x86_dummy_intr;

	/*
	 * Make pages 10 and 11 point at the PCI bus
	 */
	pages[10] = (unsigned char *)0x800a0000;
	pages[11] = (unsigned char *)0x800b0000;

	/*
	 * We want some memory for the first page.  This contains
	 * the interrupt table and stack.
	 */
	pages[0] = malloc(4 * 65536);
	pages[1] = pages[0] + 65536;

	/*
	 * BIOSes are copied to pages 12 and 13.  We want these
	 * two pages to be contiguous.
	 */
	pages[12] = pages[1] + 65536;
	pages[13] = pages[12] + 65536;

	memset(pages[0], 0, 2 * 65536);
	memset(pages[12], 0xff, 2 * 65536);
}
